# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import configparser
import ipaddress
import itertools
import json
import os
import subprocess
import sys


VENV = "/opt/stack/bifrost"
ANSIBLE = os.path.join(VENV, 'bin', 'ansible-playbook')
COMMON_ENV = {
    'VENV': VENV,
}
COMMON_PARAMS = [
    '-e', 'bifrost_venv_dir=%s' % VENV,
]
BASE = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), '..'))
PLAYBOOKS = os.path.join(BASE, 'playbooks')
DEFAULT_BRANCH = 'master'


def get_env(extra=None):
    # NOTE(dtantsur): the order here matters!
    result = os.environ.copy()
    result.update(COMMON_ENV)
    if extra:
        result.update(extra)
    return result


def log(*message, only_if=True):
    if only_if:
        print(*message, file=sys.stderr)


def process_extra_vars(extra_vars):
    for item in extra_vars:

        # argparse removes quotes, just add quotes for these vars
        if 'extra_kernel_options=' in item:
            key, value = item.split('=', 1)
            item = key + '="' + value + '"'

        if item.startswith('@'):
            # Make sure relative paths still work
            item = '@' + os.path.abspath(item[1:])

        yield ('-e', item)


def ansible(playbook, inventory, verbose=False, env=None, extra_vars=None,
            params_output_file=None, **params):
    extra = COMMON_PARAMS[:]
    if params_output_file is None:
        extra.extend(itertools.chain.from_iterable(
            ('-e', '%s=%s' % pair) for pair in params.items()
            if pair[1] is not None))
    else:
        params_output_file = os.path.abspath(params_output_file)
        log('Writing environment file', params_output_file, only_if=verbose)
        with open(params_output_file, 'wt') as output:
            json.dump({k: v for k, v in params.items() if v is not None},
                      output)
        extra.extend(['-e', '@%s' % params_output_file])

    if extra_vars:
        extra.extend(itertools.chain.from_iterable(
            process_extra_vars(extra_vars)))
    if verbose:
        extra.append('-vvvv')
    args = [ANSIBLE, playbook, '-i', inventory] + extra
    log('Calling ansible with', args, 'and environment', env, only_if=verbose)
    subprocess.check_call(args, env=get_env(env), cwd=PLAYBOOKS)


def env_setup(args):
    log('Installing dependencies and preparing an environment in', VENV,
        only_if=args.debug)
    env = get_env({'BIFROST_TRACE': str(args.debug).lower(),
                   'BIFROST_HIDE_PROMPT': 'true'})
    subprocess.check_call(["bash", "scripts/env-setup.sh"], env=env, cwd=BASE)


def get_release(release):
    if release:
        if release != DEFAULT_BRANCH and not release.startswith('stable/'):
            release = 'stable/%s' % release
        return release
    else:
        try:
            gr = configparser.ConfigParser()
            gr.read(os.path.join(BASE, '.gitreview'))
            release = gr.get('gerrit', 'defaultbranch', fallback=None)
        except (FileNotFoundError, configparser.Error):
            log('Cannot read .gitreview, falling back to defaults')
            return None
        else:
            if release and release.startswith('bugfix/'):
                log('Bugfix branch', release, 'cannot be used as a release, '
                    'falling back to defaults')
                return None
            log('Using release', release, 'detected from the checkout')
            return release


def cmd_testenv(args):
    env_setup(args)
    log('Creating', args.count, 'test node(s) with', args.memory,
        'MiB RAM and', args.disk, 'GiB of disk')

    release = get_release(args.release)

    kwargs = {}
    if release:
        kwargs['git_branch'] = release
    if args.storage_pool_path:
        kwargs['test_vm_storage_pool_path'] = os.path.abspath(
            args.storage_pool_path)

    ansible('test-bifrost-create-vm.yaml',
            inventory='inventory/localhost',
            verbose=args.debug,
            test_vm_num_nodes=args.count,
            test_vm_memory_size=args.memory,
            test_vm_disk_gib=args.disk,
            test_vm_domain_type=args.domain_type,
            test_vm_node_driver=args.driver,
            default_boot_mode=args.boot_mode or 'uefi',
            baremetal_json_file=os.path.abspath(args.inventory),
            baremetal_nodes_json=os.path.abspath(args.output),
            extra_vars=args.extra_vars,
            **kwargs)
    log('Inventory generated in', args.output)


def cmd_install(args):
    release = get_release(args.release)

    kwargs = {}
    if release:
        kwargs.update({'git_branch': release,
                       'ipa_upstream_release': release.replace('/', '-')})
    if args.dhcp_pool:
        try:
            start, end = args.dhcp_pool.split('-')
            ipaddress.ip_address(start)
            ipaddress.ip_address(end)
        except ValueError as e:
            sys.exit("Malformed --dhcp-pool, expected two IP addresses. "
                     "Error: %s" % e)
        else:
            kwargs['dhcp_pool_start'] = start
            kwargs['dhcp_pool_end'] = end

    env_setup(args)
    ansible('install.yaml',
            inventory='inventory/target',
            verbose=args.debug,
            create_ipa_image=False,
            create_image_via_dib=False,
            install_dib=True,
            network_interface=args.network_interface,
            enable_keystone=args.enable_keystone,
            enable_tls=args.enable_tls,
            generate_tls=args.enable_tls,
            noauth_mode=False,
            enabled_hardware_types=args.hardware_types,
            cleaning_disk_erase=args.cleaning_disk_erase,
            testing=args.testenv,
            download_custom_deploy_image=args.testenv,
            use_tinyipa=args.testenv,
            developer_mode=args.develop,
            enable_prometheus_exporter=args.enable_prometheus_exporter,
            default_boot_mode=args.boot_mode or 'uefi',
            enable_dhcp=not args.disable_dhcp,
            extra_vars=args.extra_vars,
            params_output_file=args.output,
            **kwargs)
    log("Ironic is installed and running, try it yourself:\n",
        " $ source %s/bin/activate\n" % VENV,
        " $ export OS_CLOUD=bifrost\n",
        " $ baremetal driver list\n"
        "See documentation for next steps")


def ensure_inside_venv():
    try:
        import oslo_config  # noqa
    except ImportError:
        sys.exit("This command must be executed inside the Bifrost virtual "
                 "environment. Try:\n"
                 f" $ source {VENV}/bin/activate\n"
                 " $ export OS_CLOUD=bifrost")


def configure_inventory(args):
    ensure_inside_venv()
    inventory = os.path.join(PLAYBOOKS, 'inventory', 'bifrost_inventory.py')
    if not args.inventory:
        os.environ['BIFROST_INVENTORY_SOURCE'] = 'ironic'
    elif os.path.exists(args.inventory):
        nodes_inventory = os.path.abspath(args.inventory)
        os.environ['BIFROST_INVENTORY_SOURCE'] = nodes_inventory
    else:
        sys.exit('Inventory file %s cannot be found' % args.inventory)
    return inventory


def cmd_enroll(args):
    inventory = configure_inventory(args)
    ansible('enroll-dynamic.yaml',
            inventory=inventory,
            verbose=args.debug,
            inspect_nodes=args.inspect,
            extra_vars=args.extra_vars)


def cmd_deploy(args):
    inventory = configure_inventory(args)
    try:
        configdrive = json.loads(args.configdrive)
    except (ValueError, TypeError):
        configdrive = args.configdrive

    extra_vars = args.extra_vars or []
    if configdrive:
        # Need to preserve JSON
        extra_vars.append(json.dumps({'deploy_config_drive': configdrive}))

    if (args.image and not args.image.startswith('file://') and not
            args.image_checksum):
        raise TypeError('An --image-checksum is required with --image '
                        'when the image is not a local file')

    ansible('deploy-dynamic.yaml',
            inventory=inventory,
            verbose=args.debug,
            deploy_image_source=args.image,
            deploy_image_type=args.image_type,
            deploy_image_checksum=args.image_checksum,
            wait_for_node_deploy=args.wait,
            extra_vars=extra_vars)


def parse_args():
    parser = argparse.ArgumentParser("Bifrost CLI")
    parser.add_argument('--debug', action='store_true',
                        help='output extensive logging')

    subparsers = parser.add_subparsers()

    testenv = subparsers.add_parser(
        'testenv', help='Prepare a virtual testing environment')
    testenv.set_defaults(func=cmd_testenv)
    testenv.add_argument('--release', default='master',
                         help='release branch to use (master, ussuri, etc), '
                              'must match the release of bifrost.')
    testenv.add_argument('--count', type=int, default=2,
                         help='number of nodes to create')
    testenv.add_argument('--memory', type=int, default=3072,
                         help='memory (in MiB) for test nodes')
    testenv.add_argument('--disk', type=int, default=10,
                         help='disk size (in GiB) for test nodes')
    testenv.add_argument('--domain-type', default='qemu',
                         help='domain type: qemu or kvm')
    testenv.add_argument('--storage-pool-path',
                         help='path to libvirt storage pool to setup')
    testenv.add_argument('--inventory', default='baremetal-inventory.json',
                         help='output file with the inventory for using '
                              'with dynamic playbooks')
    testenv.add_argument('--driver', default='ipmi',
                         choices=['ipmi', 'redfish'],
                         help='driver for testing nodes')
    boot_mode = testenv.add_mutually_exclusive_group()
    boot_mode.add_argument('--uefi', dest='boot_mode',
                           action='store_const', const='uefi',
                           help='boot testing VMs with UEFI by default')
    boot_mode.add_argument('--legacy-boot', dest='boot_mode',
                           action='store_const', const='bios',
                           help='boot testing VMs with legacy boot by default')
    testenv.add_argument('-e', '--extra-vars', action='append',
                         help='additional vars to pass to ansible')
    testenv.add_argument('-o', '--output', default='baremetal-nodes.json',
                         help='output file with the nodes information for '
                              'importing into ironic')

    install = subparsers.add_parser('install', help='Install ironic')
    install.set_defaults(func=cmd_install)
    install.add_argument('--testenv', action='store_true',
                         help='running in a virtual environment')
    install.add_argument('--develop', action='store_true',
                         help='install packages in development mode')
    install.add_argument('--dhcp-pool', metavar='START-END',
                         help='DHCP pool to use')
    install.add_argument('--release',
                         help='release branch to use (master, ussuri, etc), '
                              'the default value is determined from the '
                              '.gitreview file in the source tree')
    install.add_argument('--network-interface',
                         help='the network interface to use')
    install.add_argument('--enable-keystone', action='store_true',
                         help='enable keystone and use authentication')
    install.add_argument('--enable-tls', action='store_true',
                         help='enable self-signed TLS on API endpoints')
    install.add_argument('--hardware-types',
                         # only generic types are enabled in the simple CI
                         default='ipmi,redfish,manual-management',
                         help='a comma separated list of enabled bare metal '
                              'hardware types')
    install.add_argument('--cleaning-disk-erase',
                         action='store_true',
                         help='enable full disk cleaning between '
                              'deployments (can take a lot of time)')
    install.add_argument('--enable-prometheus-exporter', action='store_true',
                         help='Enable Ironic Prometheus Exporter')
    boot_mode = install.add_mutually_exclusive_group()
    boot_mode.add_argument('--uefi', dest='boot_mode',
                           action='store_const', const='uefi',
                           help='use UEFI boot by default')
    boot_mode.add_argument('--legacy-boot', dest='boot_mode',
                           action='store_const', const='bios',
                           help='use legacy boot (BIOS) by default')
    install.add_argument('--disable-dhcp', action='store_true',
                         help='Disable integrated dhcp server')
    install.add_argument('-e', '--extra-vars', action='append',
                         help='additional vars to pass to ansible')
    install.add_argument('-o', '--output',
                         default='baremetal-install-env.json',
                         help='output file with the ansible environment used '
                         'to install Bifrost (excluding -e arguments)')

    enroll = subparsers.add_parser(
        'enroll', help='Enroll bare metal nodes')
    enroll.set_defaults(func=cmd_enroll)
    enroll.add_argument('inventory', default='baremetal-inventory.json',
                        help='file with the inventory')
    enroll.add_argument('--inspect', action='store_true',
                        help='inspect nodes while enrolling')
    enroll.add_argument('-e', '--extra-vars', action='append',
                        help='additional vars to pass to ansible')

    deploy = subparsers.add_parser(
        'deploy', help='Deploy bare metal nodes')
    deploy.set_defaults(func=cmd_deploy)
    deploy.add_argument('inventory', nargs='?',
                        help='file with the inventory, skip to use Ironic')
    deploy.add_argument('--image', help='image URL to deploy')
    deploy.add_argument('--image-checksum',
                        help='checksum of the image to deploy')
    deploy.add_argument('--partition', action='store_const',
                        const='partition', dest='image_type',
                        help='the image is a partition image')
    deploy.add_argument('--configdrive', help='URL or JSON with a configdrive')
    deploy.add_argument('--wait', action='store_true',
                        help='wait for deployment to be finished')
    deploy.add_argument('-e', '--extra-vars', action='append',
                        help='additional vars to pass to ansible')

    args = parser.parse_args()
    if getattr(args, 'func', None) is None:
        parser.print_usage(file=sys.stderr)
        sys.exit("Bifrost CLI: error: a command is required")
    return args


def check_for_root():
    try:
        subprocess.check_call(
            '[ $(whoami) == root ] || sudo --non-interactive true',
            shell=True, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        COMMON_PARAMS.append('--ask-become-pass')


def main():
    args = parse_args()
    try:
        check_for_root()
        args.func(args)
    except Exception as exc:
        if args.debug:
            raise
        else:
            return str(exc)
    except KeyboardInterrupt:
        return 'Aborting by user request'
    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# (c) 2015, Hewlett-Packard Development Company, L.P.
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from ansible_collections.openstack.cloud.plugins.module_utils.ironic import (
    IronicModule,
    ironic_argument_spec,
)
from ansible_collections.openstack.cloud.plugins.module_utils.openstack import (  # noqa: E501
    openstack_module_kwargs,
    openstack_cloud_from_module
)
try:
    import openstack
    HAS_SDK = True
except ImportError:
    HAS_SDK = False

DOCUMENTATION = '''
---
module: os_ironic_node_info
short_description: Searches Ironic and returns node information.
extends_documentation_fragment: openstack
description:
    - Queries Ironic for a requested node and returns "node" variable with
      information about the node, or fails if the node is not found. This
      module actively prevents any passwords in the node driver_info from being
      returned.
options:
    mac:
      description:
        - unique mac address that is used to attempt to identify the host.
      required: false
      default: None
    uuid:
      description:
        - globally unique identifier (UUID) to identify the host.
      required: false
      default: None
    name:
      description:
        - unique name identifier to identify the host in Ironic.
      required: false
      default: None
    ironic_url:
      description:
        - If noauth mode is utilized, this is required to be set to the
          endpoint URL for the Ironic API.  Use with "auth" and "auth_type"
          settings set to None.
      required: false
      default: None

requirements: ["openstacksdk"]
'''

EXAMPLES = '''
# Get information about the node called "testvm1" and store in "node_info.node"
- os_ironic_node_info:
    name: "testvm1"
  register: node_info
'''


def _choose_id_value(module):
    if module.params['uuid']:
        return module.params['uuid']
    if module.params['name']:
        return module.params['name']
    return None


def main():
    argument_spec = ironic_argument_spec(  # noqa: F405
        uuid=dict(required=False),
        name=dict(required=False),
        mac=dict(required=False),
        skip_items=dict(required=False, type='list'),
    )
    module_kwargs = openstack_module_kwargs()  # noqa: F405
    module = IronicModule(argument_spec, **module_kwargs)
    module.deprecate('os_ironic_node_info is deprecated, please use '
                     'openstack.cloud.baremetal_node_info')

    if not HAS_SDK:
        module.fail_json(msg='openstacksdk is required for this module')

    try:
        sdk, cloud = openstack_cloud_from_module(module)
        if module.params['name'] or module.params['uuid']:
            server = cloud.get_machine(_choose_id_value(module))
        elif module.params['mac']:
            server = cloud.get_machine_by_mac(module.params['mac'])
        else:
            module.fail_json(msg="The worlds did not align, "
                                 "the host was not found as "
                                 "no name, uuid, or mac was "
                                 "defined.")
        if server:
            facts = dict(server)
            new_driver_info = dict()
            # Rebuild driver_info to remove any password values
            # as they will be masked.
            for key, value in facts['driver_info'].items():
                if 'password' not in key:
                    new_driver_info[key] = value
            if new_driver_info:
                facts['driver_info'] = new_driver_info

            for item in module.params['skip_items']:
                if item in facts:
                    del facts[item]

            # Remove ports and links as they are useless in the ansible
            # use context.
            if "ports" in facts:
                del facts["ports"]
            if "links" in facts:
                del facts["links"]

            nics = cloud.list_nics_for_machine(server['uuid'])
            facts['nics'] = [{'mac': nic['address']} for nic in nics]

            module.exit_json(changed=False, node=facts)

        else:
            module.fail_json(msg="node not found.")

    except openstack.exceptions.SDKException as e:
        module.fail_json(msg=e.message)


main()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_inventory
----------------------------------

Tests for `inventory` module.
"""

from unittest import mock

import openstack

from bifrost import inventory
from bifrost.tests import base


class TestBifrostInventoryUnit(base.TestCase):

    def test_inventory_preparation(self):
        (groups, hostvars) = inventory._prepare_inventory()
        self.assertIn("baremetal", groups)
        self.assertIn("localhost", groups)
        self.assertDictEqual(
            hostvars, {'127.0.0.1': {'ansible_connection': 'local'}})
        localhost_value = dict(hosts=["127.0.0.1"])
        self.assertDictEqual(localhost_value, groups['localhost'])

    @mock.patch.object(openstack, 'connect', autospec=True)
    def test__process_sdk(self, mock_sdk):
        (groups, hostvars) = inventory._prepare_inventory()
        mock_cloud = mock_sdk.return_value
        mock_cloud.list_machines.return_value = [
            {
                'driver_info': None,
                'links': [],
                'name': 'node1',
                'ports': [],
                'properties': None,
                'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
            },
        ]
        mock_cloud.get_machine.return_value = {
            'driver_info': {
                'ipmi_address': '1.2.3.4',
            },
            'links': [],
            'name': 'node1',
            'ports': [],
            'properties': {
                'cpus': 42,
            },
            'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
        }
        mock_cloud.list_nics_for_machine.return_value = [
            {
                'address': '00:11:22:33:44:55',
                'uuid': 'e2be93b5-a8f6-46a2-bec7-571b8ecf2938',
            },
        ]
        (groups, hostvars) = inventory._process_sdk(groups, hostvars)
        mock_cloud.list_machines.assert_called_once_with()
        mock_cloud.list_nics_for_machine.assert_called_once_with(
            'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45')
        self.assertIn('baremetal', groups)
        self.assertIn('hosts', groups['baremetal'])
        self.assertEqual(groups['baremetal'], {'hosts': ['node1']})
        self.assertIn('node1', hostvars)
        expected_machine = {
            'addressing_mode': 'dhcp',
            'driver_info': {
                'ipmi_address': '1.2.3.4',
            },
            'name': 'node1',
            'nics': [
                {
                    'mac': '00:11:22:33:44:55',
                },
            ],
            'properties': {
                'cpus': 42,
            },
            'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
        }
        self.assertEqual(hostvars['node1'], expected_machine)

    @mock.patch.object(openstack, 'connect', autospec=True)
    def test__process_sdk_multiple_nics(self, mock_sdk):
        (groups, hostvars) = inventory._prepare_inventory()
        mock_cloud = mock_sdk.return_value
        mock_cloud.list_machines.return_value = [
            {
                'driver_info': None,
                'links': [],
                'name': 'node1',
                'ports': [],
                'properties': None,
                'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
            },
        ]
        mock_cloud.get_machine.return_value = {
            'driver_info': {
                'ipmi_address': '1.2.3.4',
            },
            'links': [],
            'name': 'node1',
            'ports': [],
            'properties': {
                'cpus': 42,
            },
            'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
        }
        mock_cloud.list_nics_for_machine.return_value = [
            {
                'address': '00:11:22:33:44:55',
                'uuid': 'e2be93b5-a8f6-46a2-bec7-571b8ecf2938',
            },
            {
                'address': '00:11:22:33:44:56',
                'uuid': '59e8cd37-4f71-4ca1-a264-93c2ca7de0f7',
            },
        ]
        (groups, hostvars) = inventory._process_sdk(groups, hostvars)
        mock_cloud.list_machines.assert_called_once_with()
        mock_cloud.list_nics_for_machine.assert_called_once_with(
            'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45')
        self.assertIn('baremetal', groups)
        self.assertIn('hosts', groups['baremetal'])
        self.assertEqual(groups['baremetal'], {'hosts': ['node1']})
        self.assertIn('node1', hostvars)
        expected_machine = {
            'addressing_mode': 'dhcp',
            'driver_info': {
                'ipmi_address': '1.2.3.4',
            },
            'name': 'node1',
            'nics': [
                {
                    'mac': '00:11:22:33:44:55',
                },
                {
                    'mac': '00:11:22:33:44:56',
                },
            ],
            'properties': {
                'cpus': 42,
            },
            'uuid': 'f3fbf7c6-b4e9-4dd2-8ca0-c74a50f8be45',
        }
        self.assertEqual(hostvars['node1'], expected_machine)
